/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: siolook.c,v 1.3 2014/06/04 10:12:59 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sysexits.h>
#include <termios.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <err.h>
#include <sys/resource.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include "tiocm.h"

#define ERROUT(A,B) do {ex = (A); errno = (B); goto out;} while(0)
#define EXOUT(A)    do {ex = (A); goto out;} while(0)

#define B_redir 0x000001
#define B_remod 0x000020
#define B_recur 0x000040

typedef struct {        /*********************/    /*                                          */
    BLIN_flag       flags;                         /*                                           */
#   define          SIOLOOK_FOLLOW 0x0000FF        /*                     */
#   define          SIOLOOK_FORCE  0x000100        /*                     */
#   define          SIOLOOK_VIEW   0x000200        /*                      */
#   define          SIOLOOK_RUN    0x000400        /*                                    */
#   define          SIOLOOK_STATE  0x000800        /*                            */
#   define          SIOLOOK_LSTATE 0x002000        /*    stdout                       */
    babolo_opt     *opt;                           /*           */
    char           *name;                          /* path                                      */
    int             line;                          /* fd                               */
    int             set;                           /* state                 */
    int             rst;                           /* state                    */
    char           *on;
    char           *off;
    u_char          wait;                          /* line I/O wait time in ms                       */
    speed_t         baud;
    int             state;
    struct termios  term;
    struct timespec tr;                            /*                                  */
    struct timespec ts;                            /*                                   */
    struct timespec tw;                            /*                                     */
    struct rlimit   proc;                          /*                        */
} siolook_cf;

static void
/*****************************************************************************************************
 *****************************************************************************************************/
usage(int ex) {                                                                                  /****
 *****************************************************************************************************
 *****************************************************************************************************/
    fprintf( stderr
           , "siolook " VERS " @BABOLO " DATE "\n"
             "Usage: siolook [-options] line offscript onscript\n"
             "    -0 'action' - action on reset\n"
             "    -1 'action' - action on set\n"
             "    -b baud     - set baud rate\n"
             "    -c num      - soft limit of processes\n"
             "    -C num      - hard limit of processes\n"
             "    -E          - use environment variable instead of optarg\n"
             "    -f state    - action follows state line\n"
             "    -F          - force action on start\n"
             "    -L          - output state on stdout\n"
             "    -q          - quiet\n"
             "    -r state    - reset state on start\n"
             "    -s state    - set state on start\n"
             "    -t sec.nsec - display interval\n"
             "    -v[v...]    - verbose\n"
             "    -V          - view at states on start\n"
             "    -w wait     - set wait time in 0.1s units\n"
             "    -z ....     - recursive opts without environment substitute\n"
             "    -Z ....     - recursive opts with environment substitute\n"
           )
    ;
    exit(ex);
}

static char *
/*****************************************************************************************************
 *****************************************************************************************************/
redir(u_int32_t flags, babolo_opt **opt) {                                                       /****
 *****************************************************************************************************
 *****************************************************************************************************/
    babolo_parm *bp;
    u_int32_t mask;
    char *o;

    if  (flags & B_recur) {
        mask = (flags & BLIN_MASK) | Bpars_NOAL | Bpars_NONU;
        if  (flags & B_remod) mask |= Bpars_NOEN;
        if  ((o = redir(flags & B_redir, opt))) {
            if  (!(bp = babolo_getparm(mask, (u_char**)&o, (u_char*)"", 0))) {
                o = NULL;
            } else if (babolo_openopt(flags & BLIN_MASK, opt, bp->argc, (char **)bp->argv, NULL)) {
                o = NULL;
        }   }
    } else {
        if  ((o = babolo_getoptarg(opt, 0)) && (flags & B_redir)) o = getenv(o);
    }
    return(o);
}

static void
/*****************************************************************************************************
 *****************************************************************************************************/
retest(u_int32_t flags, char *o) {                                                               /****
 *****************************************************************************************************
 *****************************************************************************************************/
    for (; *o; o++) {
        if  (o[1] == ':') {
            printf("%c\t1\n", *(o++));
        } else {
            printf("%c\t0\n", *o);
    }   }
    if  (flags & B_remod) printf("\\.\n");
}

#define NOPT  0
#define OPT0  1
#define OPT1  2

static int
/*****************************************************************************************************
 *****************************************************************************************************/
retext(babolo_opt *opt) {                                                                        /****
 *****************************************************************************************************
 *****************************************************************************************************/
    int c, ex = EX_OK;
    babolo_opt *o;

    for (o = opt; o; o = o->next) {
        for (c = 0; c < 256; ++c) {
            if  (o->opts[c]) {
                if  (c < ' ') {
                    printf("\\%03o\t%u\n", c, o->opts[c] - 1);
                } else {
                    printf("%c\t%u\n", c, o->opts[c] - 1);
    }   }   }   }
    return(ex);
}

static unsigned long
/*****************************************************************************************************
 *****************************************************************************************************/
renum(int redirenv, babolo_opt **opt, int base, char c) {                                            /****
 *****************************************************************************************************
 *****************************************************************************************************/
    unsigned long ex = 0;
    char *o;

    if  (!(o = redir(redirenv, opt))) {
        ex = errno;
        ifBLIN_QW0("No -%c arg", c);
        errno = ex;
        goto out;
    }
    errno = 0;
    ex = strtoul(o, &o, 0);
    if  (errno) {
        ex = errno;
        ifBLIN_QW0("-%c strtoul", c);
        errno = ex;
        goto out;
    } else if (*o) {
        ifBLIN_QX0("-%c text", c);
        errno = ex = EPROTO;
    }
out:
    return(ex);
}

static void
/*****************************************************************************************************
 *****************************************************************************************************/
statinfo(siolook_cf *cfg, struct termios *term, int state) {                                                                      /****
 *****************************************************************************************************
 *****************************************************************************************************/
    size_t i;

    printf( " i %08X o %08X c %08X l %08X ispeed %u ospeed %u\n"
          , term->c_iflag
          , term->c_oflag
          , term->c_cflag
          , term->c_lflag
          , term->c_ispeed
          , term->c_ospeed
          )
    ;
    for (i = 0; i < NCCS; ++i) printf(" %02X", term->c_cc[i]);
    printf("\n");
    printf( "%08X %cLE %cDTR %cDSR %cRTS %cCTS %cST %cSR %cCD %cRI  %10"BLIN_D".%09ld\n"
          , state
          , (state & TIOCM_LE ) ? '+' : '-'
          , (state & TIOCM_DTR) ? '+' : '-'
          , (state & TIOCM_DSR) ? '+' : '-'
          , (state & TIOCM_RTS) ? '+' : '-'
          , (state & TIOCM_CTS) ? '+' : '-'
          , (state & TIOCM_ST ) ? '+' : '-'
          , (state & TIOCM_SR ) ? '+' : '-'
          , (state & TIOCM_CD ) ? '+' : '-'
          , (state & TIOCM_RI ) ? '+' : '-'
          , cfg->ts.tv_sec
          , cfg->ts.tv_nsec
          )
    ;
}

static void
/*****************************************************************************************************
 *****************************************************************************************************/
statinfo_v(siolook_cf *cfg) {                                                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
    struct termios term;
    int state;

    if  (0 > (ioctl(cfg->line, TIOCGETA, &term))) {
        ifBLIN_QW0("TIOCGETA %s", cfg->name);
    }
    if  (0 > (ioctl(cfg->line, TIOCMGET, &state))) {
            ifBLIN_QW0("TIOCMGET %s", cfg->name);
    }
    statinfo(cfg, &term, state);
}

static void
/*****************************************************************************************************
 *****************************************************************************************************/
statinfo_s(siolook_cf *cfg) {                                                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
    statinfo(cfg, &cfg->term, cfg->state);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
main(int argc, char *argv[], char **envp) {                                                      /****
 *****************************************************************************************************
 *****************************************************************************************************/
    int c, t, w, nxtest, redirenv, ex = EX_OK;
    char *flar = "0:1:b:c:C:Ef:FLqr:s:t:vVw:z:Z:=:?";
    siolook_cf *cfg;
    char *o;

    blin_ctl(BLIN_CTL_FSET, BLIN_BIT0);
    if  (!(cfg = calloc(1, sizeof(*cfg)))) {
        ifBLIN_QX0("Malloc failed #1");
        ERROUT(EX_OSERR, ENOMEM);
    }
    cfg->flags = BLIN_BIT0;
    cfg->opt = NULL;
    cfg->baud = 9600;
    cfg->wait = 1;
    cfg->set = 0;
    cfg->rst = 0;
    cfg->proc.rlim_cur = 6;
    cfg->proc.rlim_max = 12;
    if  (babolo_openopt(Bpars_NONU, &cfg->opt, argc, argv, flar)) {
        ifBLIN_QW0("openopt");
        EXOUT(EX_SOFTWARE);
    }
    ex = EX_USAGE;
    for (nxtest = 0, redirenv = 0; (c = babolo_getopt(&cfg->opt)) != 0; redirenv--) {
        if  (redirenv < 0) redirenv = 0;
        switch (c) {
        case '0': if  (!(o = redir(redirenv, &cfg->opt))) {
                      ifBLIN_QW0("-%c redir", c);
                      EXOUT(EX_SOFTWARE);
                  }
                ; cfg->off = o;
                ; break;
        case '1': if  (!(o = redir(redirenv, &cfg->opt))) {
                      ifBLIN_QW0("-%c redir", c);
                      EXOUT(EX_SOFTWARE);
                  }
                ; cfg->on = o;
                ; break;
        case 'b': cfg->baud = renum(redirenv, &cfg->opt, 0, c);
                ; break;
        case 'c': cfg->proc.rlim_cur = renum(redirenv, &cfg->opt, 0, c);
                ; break;
        case 'C': cfg->proc.rlim_max = renum(redirenv, &cfg->opt, 0, c);
                ; break;
        case 'E': redirenv = 2;
                ; break;
        case 'f': if  (!(o = redir(redirenv, &cfg->opt))) {
                      ifBLIN_QW0("-%c redir", c);
                      EXOUT(EX_SOFTWARE);
                  }
                ; if  (0 >= (t = babolo_testword(&tiocm, (u_char*)o))) {
                      ifBLIN_QX0("illegal flag %s", o);
                      EXOUT(EX_USAGE);
                  }
                ; cfg->flags &= ~SIOLOOK_FOLLOW;
                ; cfg->flags |= (1 << --t) & SIOLOOK_FOLLOW;
                ; break;
        case 'F': cfg->flags |= SIOLOOK_FORCE;
                ; break;
        case 'L': cfg->flags |= SIOLOOK_LSTATE;
                ; break;
        case 'q': BLIN_QUIET(cfg->flags);
                ; break;
        case 'r': if  (!(o = redir(redirenv, &cfg->opt))) {
                      ifBLIN_QW0("-%c redir", c);
                      EXOUT(EX_SOFTWARE);
                  }
                ; if  (0 >= (t = babolo_testword(&tiocm, (u_char*)o))) {
                      ifBLIN_QX0("illegal flag %s", o);
                      EXOUT(EX_USAGE);
                  }
                ; cfg->rst |= 1 << --t;
                ; break;
        case 's': if  (!(o = redir(redirenv, &cfg->opt))) {
                      ifBLIN_QW0("-%c redir", c);
                      EXOUT(EX_SOFTWARE);
                  }
                ; if  (0 >= (t = babolo_testword(&tiocm, (u_char*)o))) {
                      ifBLIN_QX0("illegal flag %s", o);
                      EXOUT(EX_USAGE);
                  }
                ; cfg->set |= 1 << --t;
                ; break;
        case 't': if  (!(o = redir(redirenv, &cfg->opt))) {
                      ifBLIN_QW0("-%c redir", c);
                      EXOUT(EX_SOFTWARE);
                  }
                ; errno = 0;
                ; cfg->tr.tv_sec = strtoul(o, &o, 10);
                ; if  (errno) {
                      ifBLIN_QW0("-t strtoul");
                      EXOUT(EX_USAGE);
                  } else if (*o == '.') {
                      int l;

                      l = strlen(++o);
                      errno = 0;
                      cfg->tr.tv_nsec = strtoul(o, &o, 10);
                      if  (errno) {
                          ifBLIN_QW0("-t .strtoul");
                          EXOUT(EX_USAGE);
                      } else if (*o) {
                          ifBLIN_QX0("-t .text");
                          ERROUT(EX_USAGE, EPROTO);
                      }
                      for (; l < 9; ++l) {
                          cfg->tr.tv_nsec *= 10;
                      }
                      for (; l > 9; --l) {
                          cfg->tr.tv_nsec /= 10;
                      }
                  } else if (*o) {
                      ifBLIN_QX0("-t text");
                      ERROUT(EX_USAGE, EPROTO);
                  }
                ; break;
        case 'v': BLIN_VERBOSE(cfg->flags);
                ; break;
        case 'V': cfg->flags |= SIOLOOK_VIEW;
                ; break;
        case 'w': cfg->wait = renum(redirenv, &cfg->opt, 0, c);
                ; break;
        case 'z': /* FALLTHROUGH */
        case 'Z': if  (!(redir(c | redirenv, &cfg->opt))) {
                      ifBLIN_QW0("-%c redir", c);
                      EXOUT(EX_SOFTWARE);
                  }
                ; break;
        case '=': if  (!(o = redir(redirenv, &cfg->opt))) {
                      ifBLIN_QW0("No -= arg");
                      ex = EX_USAGE;
                  } else {
                      switch (o[0] & 0xFF) {
                      case '0': retest(c, flar);          break;
                      case '1': retext(cfg->opt);         break;
                      case '2': babolo_dumpopt(cfg->opt); break;
                  }   }
                ; goto out;
        case '?': usage(EX_OK);
        default : ifBLIN_QX0("Illegal flag");
                ; usage(EX_USAGE);
    }   }
    ex = EX_OK;
    if  (!(cfg->name = babolo_getarg(&cfg->opt))) {
        if  (errno) {
            ifBLIN_QW0("getarg");
            EXOUT(EX_OSERR);
        } else {
            ifBLIN_QX0("getarg: sio device not defined");
            EXOUT(EX_USAGE);
    }   }
    if  (0 > (cfg->line = open(cfg->name, O_RDWR | O_NONBLOCK))) {
        ifBLIN_QW0("open line %s", cfg->name);
        EXOUT(EX_IOERR);
    }
    if (0 > (ioctl(cfg->line, TIOCGETA, &cfg->term))) {
        ifBLIN_QW0("TIOCGETA %s", cfg->name);
        EXOUT(EX_IOERR);
    }
    if  (0 > (ioctl(cfg->line, TIOCMGET, &cfg->state))) {
            ifBLIN_QW0("TIOCMGET %s", cfg->name);
    }
    if  (cfg->flags & SIOLOOK_VIEW) {
        statinfo_s(cfg);
    } else {
        ifBLIN_QO4 statinfo_v(cfg);
    }
    cfg->term.c_iflag &= ~(IGNPAR | INPCK | ISTRIP | INLCR | IGNCR | ICRNL | IXON | IXOFF | IMAXBEL);
    cfg->term.c_oflag &= ~OPOST;
    cfg->term.c_cflag &= ~(CSIZE | PARENB | CRTS_IFLOW | CDTR_IFLOW | CDSR_OFLOW | CCAR_OFLOW | MDMBUF);
    cfg->term.c_cflag |= CS8 | CREAD | CLOCAL; /* | CCTS_OFLOW | CRTS_IFLOW; */
    cfg->term.c_lflag &= ~(ECHO | ECHOCTL | ISIG | ICANON | IEXTEN | TOSTOP);
    cfg->term.c_lflag |= NOKERNINFO;
    cfg->term.c_ispeed = cfg->baud;
    cfg->term.c_ospeed = cfg->baud;
    cfg->term.c_cc[VMIN] = 0;
    cfg->term.c_cc[VTIME] = cfg->wait;
    if  (0 > (ioctl(cfg->line, TIOCSETAW, &cfg->term))) {
        ifBLIN_QW0("TIOCSETAW %s", cfg->name);
        ex = EX_IOERR;
    }
    if  (0 > (ioctl(cfg->line, TIOCMGET, &cfg->state))) {
            ifBLIN_QW0("TIOCMGET %s", cfg->name);
    }
    cfg->state &= ~cfg->rst;
    cfg->state |= cfg->set;
    if  (0 > (ioctl(cfg->line, TIOCMSET, &cfg->state))) {
            ifBLIN_QW0("TIOCMSET %s", cfg->name);
    }
    if  (0 > clock_gettime(CLOCK_UPTIME_PRECISE, &cfg->ts)) {
        ifBLIN_QW0("clock_gettime s");
        EXOUT(EX_OSERR);
    }
    while (1) {
        cfg->ts.tv_sec += cfg->tr.tv_sec;
        cfg->ts.tv_nsec += cfg->tr.tv_nsec;
        if  (cfg->ts.tv_nsec > 1000000000) {
            cfg->ts.tv_nsec -= 1000000000;
            cfg->ts.tv_sec++;
        }
        ifBLIN_QO5 statinfo_v(cfg);
        if  (0 > (ioctl(cfg->line, TIOCMGET, &cfg->state))) {
                ifBLIN_QW0("TIOCMGET %s", cfg->name);
        }
        ifBLIN_QO5 statinfo_s(cfg);
        if  (  (!(cfg->flags & SIOLOOK_RUN) && (cfg->flags & SIOLOOK_FORCE))
            || (  (cfg->flags & SIOLOOK_RUN)
               && (!(cfg->flags & SIOLOOK_STATE) != !(cfg->state & cfg->flags & SIOLOOK_FOLLOW))
            )  ) {
            if  (cfg->flags & SIOLOOK_VIEW) {
                statinfo_s(cfg);
            } else {
                ifBLIN_QO4 statinfo_v(cfg);
            }
            if  (cfg->state & cfg->flags & SIOLOOK_FOLLOW) {
                if  (cfg->on) {
                    if  (0 > (t = fork())) {
                        ifBLIN_QW0("system ON fork");
                        EXOUT(EX_OSERR);
                    } else if (!t) {
                        if  (0 > system(cfg->on)) {
                            ifBLIN_QW0("system ON: %s", cfg->on);
                            EXOUT(EX_OSERR);
                        }
                        exit(0);
                }   }
                cfg->flags |= SIOLOOK_STATE;
            } else {
                if  (cfg->off) {
                    if  (0 > (t = fork())) {
                        ifBLIN_QW0("system OFF fork");
                        EXOUT(EX_OSERR);
                    } else if (!t) {
                        if  (0 > system(cfg->off)) {
                            ifBLIN_QW0("system OFF: %s", cfg->off);
                            EXOUT(EX_OSERR);
                        }
                        exit(0);
                }   }
                cfg->flags &= ~SIOLOOK_STATE;
            }
            if  (cfg->flags & SIOLOOK_LSTATE) {
                printf("%d\n", (cfg->flags & SIOLOOK_STATE) ? 1 : 0);
                fflush(stdout);
        }   }
        cfg->flags |= SIOLOOK_RUN;
        if  ((t = waitpid(-1, &w, WNOHANG)) && WIFEXITED(w) && WEXITSTATUS(w)) {
            ifBLIN_QX1("Pid %d exit %d", t, WEXITSTATUS(w));
        }
        if  (0 > clock_gettime(CLOCK_UPTIME_PRECISE, &cfg->tw)) {
            ifBLIN_QW0("clock_gettime w");
            EXOUT(EX_OSERR);
        }
        if  (  (cfg->ts.tv_sec > cfg->tw.tv_sec)
            || ((cfg->ts.tv_sec == cfg->tw.tv_sec) && (cfg->ts.tv_nsec > cfg->tw.tv_nsec))
            ) {
            cfg->tw.tv_sec = cfg->ts.tv_sec - cfg->tw.tv_sec;
            cfg->tw.tv_nsec = cfg->ts.tv_nsec - cfg->tw.tv_nsec;
            if  (cfg->tw.tv_nsec < 0) {
                cfg->tw.tv_nsec += 1000000000;
                cfg->tw.tv_sec--;
            }
            nanosleep(&cfg->tw, NULL);
        }
    }
out:
    ifBLIN_QX2("- %d", ex);
    exit(ex);
}
